package meerkat.mixer.proofs;

import com.google.protobuf.InvalidProtocolBufferException;
import meerkat.crypto.concrete.ECElGamalEncryption;
import meerkat.crypto.concrete.Util;
import meerkat.crypto.mixnet.Mix2ZeroKnowledgeProver;
import meerkat.protobuf.Crypto;
import meerkat.protobuf.Mixing;
import org.bouncycastle.math.ec.ECPoint;
import org.factcenter.qilin.primitives.RandomOracle;
import org.factcenter.qilin.primitives.concrete.ECGroup;

import java.math.BigInteger;
import java.util.Random;


/**
 * implements of Mix2ZeroKnowledgeProver interface
 * this implementation assumes that each RerandomizableEncryptedMessage is ElGamalCiphertext
 */
public class Prover implements Mix2ZeroKnowledgeProver {

    private final ECGroup group;
    private final RandomOracle randomOracle;
    private final Random rand;
    private final ECElGamalEncryption encryptor;
    private final ECPoint g,h;
    private final BigInteger groupOrderUpperBound;
    private final ECElGamalMixParams mixParams;

    /**
     * @param rand
     * @param encryptor
     * @param randomOracle  - use for Fiat–Shamir heuristic
     */
    public Prover(Random rand, ECElGamalEncryption encryptor, RandomOracle randomOracle) {

        this.rand = rand;
        this.encryptor = encryptor;
        this.randomOracle = randomOracle;
        this.group = this.encryptor.getGroup();
        this.g = group.getGenerator();
        this.h = this.encryptor.getElGamalPK().getPK();
        this.mixParams = new ECElGamalMixParams(encryptor);
        this.groupOrderUpperBound  = group.orderUpperBound();
    }

    /**
     * @param a
     * @param b
     * @param c  - if switched then c = rerandomize(b,r2) else c = rerandomize(a,r1)
     * @param d  - if switched then d = rerandomize(a,r1) else d = rerandomize(b,r2)
     * @param switched    - trueClauseIndex
     * @param i     - column of a and c in encryption table
     * @param j     - column of b and d in encryption table
     * @param layer - row of a,b in encryption table
     * @param r1
     * @param r2
     * @return      - a valid ZKP that indeed c,d calculated as required
     * @throws InvalidProtocolBufferException
     */
    public Mixing.Mix2Proof prove(Crypto.RerandomizableEncryptedMessage a,
                                           Crypto.RerandomizableEncryptedMessage b,
                                           Crypto.RerandomizableEncryptedMessage c,
                                           Crypto.RerandomizableEncryptedMessage d,
                                           boolean switched,int i,int j, int layer,
                                           Crypto.EncryptionRandomness r1,
                                           Crypto.EncryptionRandomness r2) throws InvalidProtocolBufferException {
        Mixing.Mix2Proof first,second,third,fourth;

        ECElGamalMixParams.MixStatementWitness statement = mixParams.createProverStatement(a,b,c,d,r1,r2,switched);

        first = createOrProof(statement.getDlogStatement(0));
        second = createOrProof(statement.getDlogStatement(1));
        third = createOrProof(statement.getDlogStatement(2));
        fourth = createOrProof(statement.getDlogStatement(3));

        Mixing.ZeroKnowledgeProof.Location location = Mixing.ZeroKnowledgeProof.Location.newBuilder()
                .setI(i)
                .setJ(j)
                .setLayer(layer)
                .build();

        Mixing.ZeroKnowledgeProof result = Mixing.ZeroKnowledgeProof.newBuilder()
                .setFirst(first)
                .setSecond(second)
                .setThird(third)
                .setFourth(fourth)
                .setLocation(location)
                .build();
        return  result;
    }


    /**
     * Fiat–Shamir heuristic
     * @param input - protobuf contains all parameters from the first step of the current proof
     * @param randomOracle
     * @return randomOracle.hash(input)
     */
    public static BigInteger hash(Mixing.Mix2Proof.FirstMessage input, RandomOracle randomOracle) {
        byte[] arr = input.toByteArray();
        return new BigInteger(1,randomOracle.hash(arr,arr.length));
    }



    Mixing.Mix2Proof.DlogProof.FirstMessage createDlogProof(ECElGamalMixParams.DlogStatement statement, ECElGamalMixParams.DlogStatementWitness witness) {

        BigInteger r = encryptor.generateRandomExponent(rand);
        ECPoint gr = group.multiply(statement.g, r);
        ECPoint hr = group.multiply(statement.h, r);

        Mixing.Mix2Proof.DlogProof.FirstMessage firstMessage = Mixing.Mix2Proof.DlogProof.FirstMessage.newBuilder()
                .setGr(group.encode(gr))
                .setHr(group.encode(hr))
                .build();

        BigInteger challenge = Prover.hash()
    }




    /**
     * Generate a ZK proof that there exists x s.t (g ^ x == a and h ^ x == b) or (g' ^ x == a and h' ^ x == b).
     *
     * For each clause of the disjunction, we use the following sigma-protocol for DLOG equality (i.e. log_{g}(a)==log_{g}(b)):
     * <ol>
     *     <li>Prover chooses a random r, and sends g^r, h^r </li>
     *     <li>Verifier chooses a random c and sends c</li>
     *     <li>Prover computes </li>
     * </ol>
     *
     *
     * @param orStatement
     * @return ZKP OrProof:
     *         assuming DLog is hard in this.group then that proves x is known for the meerkat.mixer.proofs
     */
    private Mixing.ZeroKnowledgeProof.OrProof createOrProof(ECElGamalMixParams.OrProverStatement orStatement) {

        ECPoint g1 = orStatement.g1;
        ECPoint h1 = orStatement.h1;
        ECPoint g2 = orStatement.g2;
        ECPoint h2 = orStatement.h2;

        ECPoint g1Tag  = orStatement.g1Tag;
        ECPoint h1Tag  = orStatement.h1Tag;
        ECPoint g2Tag  = orStatement.g2Tag;
        ECPoint h2Tag  = orStatement.h2Tag;

        // Randomness for the ZK proof
        BigInteger r = encryptor.generateRandomExponent(rand);


        BigInteger c1,c2,z,zTag;
        ECPoint u,v,uTag,vTag;
        Mixing.ZeroKnowledgeProof.OrProof.FirstMessage firstMessage;


        switch (orStatement.trueClauseIndex) {
            case left:
                c2 = encryptor.generateRandomExponent(rand);
                zTag = encryptor.generateRandomExponent(rand);
                //step 1
                u = group.multiply(g1, r);
                v = group.multiply(g2, r);
                uTag = group.add(group.multiply(g1Tag, zTag), group.negate(group.multiply(h1Tag, c2)));
                vTag = group.add(group.multiply(g2Tag, zTag), group.negate(group.multiply(h2Tag, c2)));
                //step 2
                // c1 = (hash(input + step1) + group size - c2)% group size
                firstMessage =
                        Mixing.ZeroKnowledgeProof.OrProof.FirstMessage.newBuilder()
                                .setG1(orStatement.g1Encoded)
                                .setH1(orStatement.h1Encoded)
                                .setG2(orStatement.g2Encoded)
                                .setH2(orStatement.h2Encoded)
                                .setG1Tag(orStatement.g1TagEncoded)
                                .setH1Tag(orStatement.h1TagEncoded)
                                .setG2Tag(orStatement.g2TagEncoded)
                                .setH2Tag(orStatement.h2TagEncoded)
                                .setU(encryptor.encodeElement(u))
                                .setV(encryptor.encodeElement(v))
                                .setUTag(encryptor.encodeElement(uTag))
                                .setVTag(encryptor.encodeElement(vTag))
                                .build();
                c1 = hash(firstMessage,randomOracle).add(group.orderUpperBound().subtract(c2)).mod(groupOrderUpperBound);
                //step 3
                //z = (r +  c1 * x) % group size;
                z = r.add(c1.multiply(encryptor.extractRandomness(orStatement.x))).mod(groupOrderUpperBound);
                break;
            case right:
                c1 = encryptor.generateRandomExponent(rand);
                z = encryptor.generateRandomExponent(rand);
                //step 1
                uTag = group.multiply(g1Tag, r);
                vTag = group.multiply(g2Tag, r);
                u = group.add(group.multiply(g1, z), group.negate(group.multiply(h1, c1)));
                v = group.add(group.multiply(g2, z), group.negate(group.multiply(h2, c1)));
                //step 2
                // c1 = (hash(input + step1) + group size - c1)% group size
                firstMessage =
                        Mixing.ZeroKnowledgeProof.OrProof.FirstMessage.newBuilder()
                                .setG1(orStatement.g1Encoded)
                                .setH1(orStatement.h1Encoded)
                                .setG2(orStatement.g2Encoded)
                                .setH2(orStatement.h2Encoded)
                                .setG1Tag(orStatement.g1TagEncoded)
                                .setH1Tag(orStatement.h1TagEncoded)
                                .setG2Tag(orStatement.g2TagEncoded)
                                .setH2Tag(orStatement.h2TagEncoded)
                                .setU(encryptor.encodeElement(u))
                                .setV(encryptor.encodeElement(v))
                                .setUTag(encryptor.encodeElement(uTag))
                                .setVTag(encryptor.encodeElement(vTag))
                                .build();
                c2 = hash(firstMessage,randomOracle).add(group.orderUpperBound().subtract(c1)).mod(groupOrderUpperBound);
                //step 3
                //zTag = (r +  c2 * x) % group size;
                zTag = r.add(c2.multiply(encryptor.extractRandomness(orStatement.x))).mod(groupOrderUpperBound);
                break;
            default:
                return null;
        }


        return Mixing.ZeroKnowledgeProof.OrProof.newBuilder()
                .setG1(firstMessage.getG1())
                .setH1(firstMessage.getH1())
                .setG2(firstMessage.getG2())
                .setH2(firstMessage.getH2())
                .setG1Tag(firstMessage.getG1())
                .setH1Tag(firstMessage.getH1Tag())
                .setG2Tag(firstMessage.getG2Tag())
                .setH2Tag(firstMessage.getH2Tag())
                .setU(firstMessage.getU())
                .setV(firstMessage.getV())
                .setUTag(firstMessage.getUTag())
                .setVTag(firstMessage.getVTag())
                .setC1(Util.encodeBigInteger(c1))
                .setC2(Util.encodeBigInteger(c2))
                .setZ(Util.encodeBigInteger(z))
                .setZTag(Util.encodeBigInteger(zTag))
                .build();
    }
}

