package meerkat.mixer.proofs;

import com.google.protobuf.Message;

import java.math.BigInteger;

/**
 * Disjunction of 2 Sigma protocol statements
 */
public class SigmaProtocolOr2 {
    abstract public class Prover implements SigmaProtocol.Prover {
        final SigmaProtocol.Prover prover;
        final SigmaProtocol.Simulator simulator;
        final int proverIdx;

        abstract protected Message buildConcatenatedFirstMessage(Message... firstMessages);
        abstract protected Message buildConcatenatedFinalMessage(BigInteger firstChallenge, Message... finalMessages);

        /**
         * Subtract two challenges in an appropriate way (e.g., mod group order)
         * @param c1
         * @param c2
         * @return
         */
        abstract protected BigInteger subtractChallenge(BigInteger c1, BigInteger c2);

        /**
         *
         * @param prover
         * @param simulator
         * @param proverIdx Which index should the prover be inserted at (0 means first, 1 means second)
         */
        public Prover(SigmaProtocol.Prover prover, SigmaProtocol.Simulator simulator, int proverIdx) {
            this.prover = prover;
            this.simulator = simulator;
            this.proverIdx = proverIdx;
            assert (proverIdx >= 0 && proverIdx < 2);
        }

        @Override
        public Message getFirstMessage() {
            if (proverIdx == 0) {
                return buildConcatenatedFirstMessage(prover.getFirstMessage(), simulator.getFirstMessage());
            } else {
                return buildConcatenatedFirstMessage(simulator.getFirstMessage(), prover.getFirstMessage());
            }
        }


        @Override
        public Message getFinalMessage(BigInteger challenge) {
            BigInteger realchallenge = subtractChallenge(challenge, simulator.getChallenge());
            if (proverIdx == 0) {
                return buildConcatenatedFinalMessage(realchallenge, prover.getFinalMessage(realchallenge), simulator.getFinalMessage());
            } else {
                return buildConcatenatedFinalMessage(simulator.getChallenge(), simulator.getFinalMessage(), prover.getFinalMessage(realchallenge));
            }
        }
    }

    abstract public class Verifier implements SigmaProtocol.Verifier {
        final SigmaProtocol.Verifier[] verifiers;

        abstract protected Message getFirstMessage(Message concatenated, int verifier);
        abstract protected BigInteger getFirstChallenge(Message concatenated);
        abstract protected Message getFinalMessage(Message concatenated, int verifier);


        public Verifier(SigmaProtocol.Verifier... verifiers) {
            this.verifiers = verifiers;
        }

        public boolean verify(Message firstMessage, BigInteger challenge, Message finalMessage) {

            BigInteger firstChallenge = getFirstChallenge(finalMessage);
            BigInteger secondChallenge = challenge.subtract(firstChallenge);
            return verifiers[0].verify(getFirstMessage(firstMessage, 0),  firstChallenge, getFinalMessage(finalMessage, 0)) &
                    verifiers[0].verify(getFirstMessage(firstMessage, 1),  secondChallenge, getFinalMessage(finalMessage, 1));
        }
    }

}
