package meerkat.mixer.proofs;

import com.google.protobuf.InvalidProtocolBufferException;
import meerkat.crypto.concrete.ECElGamalEncryption;
import meerkat.crypto.concrete.Util;
import meerkat.crypto.mixnet.Mix2ZeroKnowledgeVerifier;
import meerkat.protobuf.ConcreteCrypto.GroupElement;
import meerkat.protobuf.Crypto;
import meerkat.protobuf.Mixing;
import org.bouncycastle.math.ec.ECPoint;
import org.factcenter.qilin.primitives.RandomOracle;
import org.factcenter.qilin.primitives.concrete.ECGroup;

import java.math.BigInteger;

/**
 * implements Mix2ZeroKnowledgeVerifier
 */
public class Verifier implements Mix2ZeroKnowledgeVerifier {


    private final ECGroup group;
    private final RandomOracle randomOracle;
    private final ECPoint g,h;
    private final ECElGamalMixParams mixParams;
    private final ZeroKnowledgeOrProofParser parser;

    /**
     * constructor
     * @param encryptor should be as the encryptor used by meerkat.mixer.proofs
     * @param randomOracle should be as the random oracle used by meerkat.mixer.proofs
     */
    public Verifier(ECElGamalEncryption encryptor, RandomOracle randomOracle) {
        this.group = encryptor.getGroup();
        this.g = group.getGenerator();
        this.h = encryptor.getElGamalPK().getPK();
        this.randomOracle = randomOracle;
        this.mixParams = new ECElGamalMixParams(encryptor);
        this.parser = new ZeroKnowledgeOrProofParser(group);
    }

    /**
     * verify zero knowledge proof
     * @param in1
     * @param in2
     * @param out1
     * @param out2
     * @param proof out1 = rerandomize(in1) && out2 = rerandomize(in2)
     *              ||
     *              out1 = rerandomize(in2) && out2 = rerandomize(in1)
     * @return  true iff all 4 or proofs are valid
     * @throws InvalidProtocolBufferException
     */
    public boolean verify(Crypto.RerandomizableEncryptedMessage in1,
                          Crypto.RerandomizableEncryptedMessage in2,
                          Crypto.RerandomizableEncryptedMessage out1,
                          Crypto.RerandomizableEncryptedMessage out2,
                          Mixing.ZeroKnowledgeProof proof) throws InvalidProtocolBufferException {
        ECElGamalMixParams.Mix2Statement statement = mixParams.createStatement(in1,in2,out1,out2);
        return verifyFiatShamirOrProof(statement, proof.getFirst(), proof.getSecond());

    }


    /**
     * Verify a Schnorr proof of discrete log equality, given the entire transcript of the ZKP.
     * @param statement
     * @param firstMessage
     * @param challenge
     * @param finalMessage
     * @return
     */
    boolean verifyInteractiveDlogProof(ECElGamalMixParams.DlogStatement statement,
                                       Mixing.ZeroKnowledgeProof.DlogProof.FirstMessage firstMessage,
                                       BigInteger challenge,
                                       Mixing.ZeroKnowledgeProof.DlogProof.FinalMessage finalMessage) {
        GroupElement grEncoded = firstMessage.getGr();
        ECPoint gr = group.decode(grEncoded.toByteArray());

        GroupElement hrEncoded = firstMessage.getHr();
        ECPoint hr = group.decode(hrEncoded.toByteArray());

        BigInteger xcr = Util.decodeBigInteger(finalMessage.getXcr());

        boolean gGood = group.add(gr, group.multiply(statement.a,challenge)).equals(group.multiply(statement.g,xcr));
        boolean hGood = group.add(hr, group.multiply(statement.b,challenge)).equals(group.multiply(statement.h,xcr));
        return gGood && hGood;
    }

    /**
     * Verify a conjuction of two Dlog Proofs, given a complete transcript of the ZKP
     * @param statement
     * @param firstMessage
     * @param challenge
     * @param finalMessage
     * @return
     */
    boolean verifyInteractiveAndProof(ECElGamalMixParams.AndStatement statement,
                                      Mixing.ZeroKnowledgeProof.AndProof.FirstMessage firstMessage,
                                      BigInteger challenge,
                                      Mixing.ZeroKnowledgeProof.AndProof.FinalMessage finalMessage) {
        return verifyInteractiveDlogProof(statement.clauses[0], firstMessage.getClause0(), challenge, finalMessage.getClause0()) &&
                verifyInteractiveDlogProof(statement.clauses[1], firstMessage.getClause1(), challenge, finalMessage.getClause1());
    }

    /**
     * Verify a disjunction of two Dlog-equality statement conjuctions, given the entire transcript of the ZKP.
     * @param statement
     * @param firstMessage
     * @param challenge
     * @param finalMessage
     * @return
     */
    boolean verifyInteractiveMix2Proof(ECElGamalMixParams.Mix2Statement statement,
                                       Mixing.ZeroKnowledgeProof.OrProof.FirstMessage firstMessage,
                                       BigInteger challenge,
                                       Mixing.ZeroKnowledgeProof.OrProof.FinalMessage finalMessage) {
        BigInteger c0 = Util.decodeBigInteger(finalMessage.getC0());
        BigInteger c1 = challenge.subtract(c0).mod(group.orderUpperBound());

        return verifyInteractiveAndProof(statement.clauses[0], firstMessage.getClause0(), c0, finalMessage.getClause0()) &&
                verifyInteractiveAndProof(statement.clauses[1], firstMessage.getClause1(), c1, finalMessage.getClause1());
    }


    boolean verifyFiatShamirOrProof(ECElGamalMixParams.Mix2Statement statement, Mixing.ZeroKnowledgeProof.OrProof.FirstMessage firstMessage,
                                    Mixing.ZeroKnowledgeProof.OrProof.FinalMessage finalMessage) {
        BigInteger challenge = Prover.hash(firstMessage, randomOracle).mod(group.orderUpperBound());
        return verifyInteractiveMix2Proof(statement, firstMessage, challenge, finalMessage);
    }
}
